#include <iostream>
#include <fstream>
#include <list>
#include <vector>
#include <map>
#include <string>
#include <set>
#include <ext/hash_map>
#include <algorithm>
#include <sys/time.h>
#include <sys/resource.h>

using namespace std;
using namespace __gnu_cxx;
double factorial[10000000];

// maths
double log10(int x)
{
 return log((double)x)/log(10.0); 
}

double fac(int x) {  
  if (x < 2) return 0;
  if (!factorial[x])
  {       
   factorial[x] = log10(x)+fac(x-1);
  }      
  return factorial[x];
}       

double C(int N, int r) {
 return fac(N)-(fac(r)+fac(N-r));
}         

// Comparison routine for clusters
bool clustersCmp(const pair<int, float> &s1, const pair<int, float> &s2)
{
  return s1.second > s2.second;
}

// Comparison routine for strings
struct strCmp {
 bool operator()(const string &s1, const string &s2 ) const 
 {
  return s1.compare(s2) < 0;
 }
};

typedef hash_map<int, float> adj_list;
typedef hash_map<int, adj_list> interaction_map;
typedef vector<set<int> > int_set_vector;
typedef set<int> int_set;

// Compute network density of combined cluster
float ICD(set<int> &s1, set<int> &s2, interaction_map &interactions)
{
 float density = 0;
 set<int> sd1, sd2;
 insert_iterator<set<int> > sd1_ins(sd1, sd1.begin()), sd2_ins(sd2, sd2.begin());
 set_difference(s1.begin(), s1.end(), s2.begin(), s2.end(), sd1_ins);
 set_difference(s2.begin(), s2.end(), s1.begin(), s1.end(), sd2_ins);

 // cout << n << " - ";
 adj_list::iterator at;
 for (set<int>::iterator it = sd1.begin(); it != sd1.end(); it++)
 {
  adj_list &alist = interactions.find(*it)->second;
  for (adj_list::iterator at = alist.begin(); at != alist.end(); at++)
    if (sd2.find(at->first) != sd2.end())
     density ++;//+= at->second;
 }
 return density/(sd1.size()*sd2.size());
}

float getOverlap(set<int> &s1, set<int> &s2)
{
 set<int> sd;
 insert_iterator<set<int> > sd_ins(sd, sd.begin());
 set_intersection(s1.begin(), s1.end(), s2.begin(), s2.end(), sd_ins);
 return ((float)(sd.size()*sd.size()))/(s1.size()*s2.size());
}

float getOverlapProb(set<int> &s1, set<int> &s2, int N)
{
 set<int> sd;
 insert_iterator<set<int> > sd_ins(sd, sd.begin());
 set_intersection(s1.begin(), s1.end(), s2.begin(), s2.end(), sd_ins);
 double pval = C(s1.size(), sd.size()) + C(N - s1.size(), s2.size() - sd.size()) - C(N, s2.size());;
// cout <<  pval << endl;
 return pval;
}

// Compute network density of combined cluster
pair<int, float> getFSWeight(set<int> s, interaction_map &interactions)
{
 int links = 0;
 float weight = 0;
 for (set<int>::iterator st = s.begin(); st != s.end(); st++)
 {
  adj_list &alist = interactions.find(*st)->second;
  for (adj_list::iterator it = alist.begin(); it != alist.end(); it++)
  {
   if (s.find(it->first) != s.end())
   {
    weight += it->second;
    links ++;
   }
  }
 }
 return pair<int, float>(links/2, weight/2);
}

// Tokenize string into a vector of substrings separated by delim
vector<string> split(const string& s, char delim)
{
 vector<string> tokens;
 string::const_iterator it, last = s.begin();
 for(it = s.begin(); it != s.end(); it++)
 {
  if (*it == delim)
  {
   tokens.push_back(string(last, it));
   last = it+1;
  }
 }
 if (s.length()) tokens.push_back(string(last, s.end()));
 return tokens;
}


// maximal clique finding (Tomita et al.)
void _EXPAND(
 int_set &members,
 int_set &subgraph,
 int_set &candidates,
 interaction_map &interactions,
 int mincliquesize,
 void *params,
 void (*fct) (void*, int_set&)
)
{
 if (!subgraph.size())
 {
  if (members.size() >= mincliquesize)
   (*fct)(params, members);
//   cliques.push_back(members);
  return;
 }
 int u = -1;
 int max = 0;
 for(set<int>::iterator it = subgraph.begin(); it != subgraph.end(); it++)
 {
  int intersect = 0;
  adj_list &alist = interactions.find(*it)->second;
  if (alist.size() <= max) continue;
  for(adj_list::iterator at = alist.begin(); at != alist.end(); at++)
  {
   if (candidates.find(at->first) != candidates.end()) 
    intersect++;
  }
  if (u == -1 || intersect > max)
  {
   max = intersect;
   u = *it;
  }   	
 }

 adj_list &alist = interactions.find(u)->second;
 set<int> diff;
 for (set<int>::iterator it = candidates.begin(); it != candidates.end(); it++)
 {
  if (alist.find(*it) != alist.end()) continue;
  diff.insert(*it);
 }
 while(diff.size())
 {
  set<int>::iterator it = diff.begin();
  adj_list &qlist = interactions.find(*it)->second;
  set<int> rq;
  for(adj_list::iterator qt = qlist.begin(); qt != qlist.end(); qt++)
   rq.insert(qt->first);

  members.insert(*it);
  set<int> newsubgraph;
  insert_iterator<set<int> > subgraph_ins(newsubgraph, newsubgraph.begin());
  set_intersection(subgraph.begin(), subgraph.end(), rq.begin(), rq.end(), subgraph_ins);
  set<int> newcandidates;
  insert_iterator<set<int> > candidates_ins(newcandidates, newcandidates.begin());
  set_intersection(candidates.begin(), candidates.end(), rq.begin(), rq.end(), candidates_ins);
  _EXPAND(members, newsubgraph, newcandidates, interactions, mincliquesize, params, fct);
  members.erase(*it);
  candidates.erase(*it);
  diff.erase(it);
 }
  
}
 
void CLIQUES(
 interaction_map &interactions,
 int mincliquesize,
 void *params,
 void (*fct) (void*, int_set&)
)
{
 set<int> members;   
 set<int> vertices;
 set<int> candidates;

 time_t starttime = time(NULL);
 
 for(interaction_map::iterator it = interactions.begin(); it != interactions.end(); it++)
 {
  if (it->second.size() < mincliquesize-1) continue;
  vertices.insert(it->first);
  candidates.insert(it->first);
 }
 _EXPAND(members, vertices, candidates, interactions, mincliquesize, params, fct); 

 time_t endtime = time(NULL);
 cout << "Clique finding completed in " << (endtime - starttime) << " second." << endl;
}


// Heuristic Clique Finding
void hCLIQUES(
 interaction_map &interactions,
 int mincliquesize,
 void *params,
 void (*fct) (void*, int_set&)
)
{
 
 time_t starttime = time(NULL);
 for(interaction_map::iterator it = interactions.begin(); it != interactions.end(); it++)
 {
  int a = it->first;
  adj_list &alist = it->second;
 
  list<pair<int, float> > nodes;
  for(adj_list::iterator at = alist.begin(); at != alist.end(); at++)
  {
   adj_list &alist1 = interactions.find(at->first)->second;
   nodes.push_back(pair<int, float>(at->first, (float)alist1.size()));
  } 
  nodes.sort(clustersCmp);

  set<int> newclique;
  newclique.insert(a);

  for(list<pair<int, float> >::iterator it = nodes.begin(); it != nodes.end(); it++)
  {
   int b = it->first;
   adj_list &alist = interactions.find(b)->second;
   int flag = 0;
   for (set<int>::iterator st = newclique.begin(); st != newclique.end(); st++)
   {
    if (alist.find(*st) == alist.end())
    {
     flag = 1;
     break;
    }
   }
   if (!flag)
   {
    newclique.insert(b);
//    it = --nodes.erase(it);
   }
  }
  if (newclique.size() >= mincliquesize)
    (*fct)(params, newclique);
 }
 time_t endtime = time(NULL);
 cout << "Clique finding completed in " << (endtime - starttime) << " second." << endl;

}

void filterclusters(int_set_vector &clusters, int index, interaction_map &interactions)
{ 
 list<pair<int, float> > sortedclusters;

 time_t starttime = time(NULL);

 int i = 0;
 int *owners = new int[index];
 memset(owners, 0, sizeof(int)*index);

 for(int_set_vector::iterator it = clusters.begin(); it != clusters.end(); it++)
 {
  pair<int, float> fsweight = getFSWeight(*it, interactions);
  float avgfsweight = fsweight.first?(fsweight.second/fsweight.first):0; 
  sortedclusters.push_back(pair<int, float>(i++, avgfsweight));

  for(set<int>::iterator st = it->begin(); st != it->end(); st++)
    owners[*st]++;
 }
 sortedclusters.sort(clustersCmp); 

 int_set_vector tempclusters;
 while(!sortedclusters.empty())
 {
  pair<int, float> s = sortedclusters.front();
  sortedclusters.pop_front();

  set<int> &s1 = clusters[s.first];

  // check for overlaps
  vector<int> removals;
  int overlap = 0;
  for(set<int>::iterator st = s1.begin(); st != s1.end(); st++)
  {
   if (owners[*st] == 0)
   {
    removals.push_back(*st);
    continue;
   }
   if (owners[*st] > 1)
   {
    overlap = 1;
    break;
   }
  }
  for(vector<int>::iterator it = removals.begin(); it != removals.end(); it++)
   s1.erase(*it);

  if (!s1.size()) continue;
  if (!overlap)
  {
   tempclusters.push_back(s1);
   continue;
  }

  for (list<pair<int, float> >::iterator it = sortedclusters.begin(); it != sortedclusters.end(); it++)
  {
   set<int> &s2 = clusters[it->first];

   // find intersection
   set<int> c;			// intersection
   set_intersection(s1.begin(), s1.end(), s2.begin(), s2.end(), inserter(c, c.begin()));
   if (!c.empty())
   {
    set<int> a;			// s1 - c
    set_difference(s1.begin(), s1.end(), c.begin(), c.end(), inserter(a, a.begin()));
    set<int> b;			// s2 - c
    set_difference(s2.begin(), s2.end(), c.begin(), c.end(), inserter(b, b.begin()));

    pair<int, float> fsweighta = getFSWeight(a, interactions);
    pair<int, float> fsweightb = getFSWeight(b, interactions);
    pair<int, float> fsweightac = getFSWeight(s1, interactions);
    pair<int, float> fsweightbc = getFSWeight(s2, interactions);        
    
    if ((fsweightbc.second + fsweighta.second)/(fsweightbc.first + fsweighta.first) 
      > (fsweightac.second + fsweightb.second)/(fsweightac.first + fsweightb.first))
    {
     s1 = a;	// remove c from s1
     if (s1.size() < 1)
     {
      s2.insert(s1.begin(), s1.end());
      s1.clear();
     }
    }else{
     s2 = b;	// remove c from s2
     if (s2.size() < 1)
     {
      s1.insert(s2.begin(), s2.end());
      s2.clear();
     }
    }

   }
  }
  if (s1.size())
  {
    for (set<int>::iterator rt = s1.begin(); rt != s1.end(); rt++)
     owners[*rt] = 0;
   tempclusters.push_back(s1);
  }
 }
 delete[] owners;

 clusters = tempclusters;

 time_t endtime = time(NULL);
 cout << "Clique filtering completed in " << (endtime - starttime) << " second." << endl;
 cout << "Filtered to form " << clusters.size() << " clusters." << endl;
}

void insert_new_clique(void *params, set<int> &members)
{
 int_set_vector &clusters = *((int_set_vector*)((void**)params)[0]);
 clusters.push_back(members);
}

void insert_new_partialclique(void* params, set<int> &members)
{
 int_set_vector &clusters = *((int_set_vector*)((void**)params)[0]);
 list<set<int> > &partialcliques = *((list<set<int> >*)((void**)params)[1]);
 int &merges = *((int*)((void**)params)[2]);

 set<int> newset;
 for(set<int>::iterator st = members.begin(); st != members.end(); st++)
 {
  set<int> &member = clusters[*st];
  newset.insert(member.begin(), member.end());
 }
// if (newset.size() > 90) { return; }
  
 bool insert = 1;
 for(list<set<int> >::iterator rt = partialcliques.begin(); rt != partialcliques.end(); rt++)
 {
  set<int> sd;
  insert_iterator<set<int> > sd_ins(sd, sd.begin());
  if (rt->size()<newset.size())
  {
   set_difference(rt->begin(), rt->end(), newset.begin(), newset.end(), sd_ins);
   if (!sd.size())
    rt = --partialcliques.erase(rt);
  }
  else
  {
   set_difference(newset.begin(), newset.end(), rt->begin(), rt->end(), sd_ins);
   if (!sd.size())
   {
    insert = 0;
    break;
   }
  }
 }
 if (insert)
 {
  partialcliques.push_back(newset);
  merges++;
 }
}

bool findpartialcliques(
 int_set_vector &clusters, 
 interaction_map &interactions, 
 float ICDthreshold)
{
 time_t starttime = time(NULL);
 
 // compute ICDs
 int numlinks = 0;
 interaction_map ICD_interactions;

 // find ownership
 int i = 1;
 int *owner = new int[interactions.size()];
 memset(owner, 0, sizeof(int)*interactions.size());
 for (int_set_vector::iterator it = clusters.begin(); it != clusters.end(); it++)
 {
  for (set<int>::iterator st=it->begin(); st != it->end(); st++)
   owner[*st] = i;
  i++;
 }
 i = 0;
 for (int_set_vector::iterator it = clusters.begin(); it != clusters.end(); it++)
 {
  bool found = 0;

  // find all neighbours;
  set<int> neighbourclusters;
  for (set<int>::iterator st=it->begin(); st != it->end(); st++)
  {
   adj_list &alist = interactions.find(*st)->second;
   for (adj_list::iterator at = alist.begin(); at != alist.end(); at++)
   {
    if (owner[at->first] > 0)
     neighbourclusters.insert(owner[at->first]-1);
   }
  }

  for (set<int>::iterator st = neighbourclusters.begin(); st != neighbourclusters.end(); st++)
  {
   if (*st < i) continue;
   float ICD_score = ICD(*it, clusters[*st], interactions);
   if (ICD_score >= ICDthreshold)
   {
    ICD_interactions[i][*st] = ICD_interactions[*st][i] = 1;
    numlinks ++;
    found = 1;
   }
  }
  i++;
 }
 delete[] owner;
 cout << "Found " << numlinks << " edges between " << ICD_interactions.size() << " cliques." << endl;

 time_t endtime = time(NULL);
 cout << "Edge finding completed in " << (endtime - starttime) << " second." << endl;

 if (!numlinks) return 0;

 list <int_set> partialcliques;
 for(int_set_vector::iterator it = clusters.begin(); it != clusters.end(); it++)
  partialcliques.push_back(*it);

 // find partial cliques
 int merges = 0;
 void* params[3];
 params[0] = &clusters;
 params[1] = &partialcliques;
 params[2] = &merges;
 CLIQUES(ICD_interactions, 2, (void*)&params, insert_new_partialclique);
 if (!merges) return 0;
 
 clusters.clear();
 clusters.insert(clusters.begin(), partialcliques.begin(), partialcliques.end());
 cout << "Merged to form " << clusters.size() << " partial cliques. - " << merges << endl;
 return 1; 
}

void printreport(int_set_vector &clusters)
{
 hash_map<int, bool> test;
 for(int_set_vector::iterator it = clusters.begin(); it != clusters.end(); it++)
 {
  for(set<int>::iterator st = it->begin(); st != it->end(); st++)
   test[*st] = 1;
 } 
 cout << "Found " << clusters.size() << " cliques among " << test.size() << " proteins"  << endl;
}

int main(int argc, char* argv[])
{ 
/*
 double **R = new double*[100];
 for (int i=0; i<100; i++)
 {
  R[i] = new double[100];
 }
 R[3][4] = R[3][5] = R[3][6] = 1;
 R[4][3] = R[4][5] = R[4][6] = 1;
 R[5][3] = R[5][4] = R[5][6] = 1;
 R[6][3] = R[6][4] = R[6][5] = 1;
 R[7][8] = R[8][7] = 1;
 set <int> clique;
 biclique(R, 100, clique);
 return 0;
*/ 
 srand(time(NULL));
 memset(factorial, 0, sizeof(factorial));

 if (argc < 2) return -1;

 float ICDthreshold = 0.1;
 if (argc > 2)
  ICDthreshold = atof(argv[2]); 

 int mincliquesize = 4;
 if (argc > 3)
  mincliquesize = atoi(argv[3]); 
 
 // read interaction file
 int index_a, index_b, index = 0;
 vector<string> protnames;
 map<string, int, strCmp> protindex;
 map<string, int, strCmp>::iterator mt;
 interaction_map interactions;
 string a, b;
 string buffer;
 ifstream intfile(argv[1]);
 while(!intfile.eof())
 {
  getline(intfile, buffer);
  vector<string> tokens = split(buffer, '\t');
  if (tokens.size() < 3) break;
  a = tokens.at(0);
  b = tokens.at(1);
  float weight = atof(tokens.at(2).c_str());
// cout << tokens.at(0) << "\t" << tokens.at(1) << endl;

  mt = protindex.find(a);
  if (mt != protindex.end()) index_a = (*mt).second;
  else
  {
   index_a = index++;
   protindex[a] = index_a;
   protnames.insert(protnames.end(), a);
  }
  mt = protindex.find(b);
  if (mt != protindex.end()) index_b = (*mt).second;
  else
  {
   index_b = index++;
   protindex[b] = index_b;
   protnames.insert(protnames.end(), b);
  }

  // add to interactions
  if (index_a != index_b) interactions[index_a][index_b] = interactions[index_b][index_a] = weight;
//  cout << index_a << "\t" << index_b << endl;
 } 
 intfile.close();

 // find cliques
 int_set_vector clusters;
 void *params[2];
 params[0] = &clusters;
 params[1] = &interactions;

 struct timeval tim;
 struct rusage ru;
 getrusage(RUSAGE_SELF, &ru);
 tim = ru.ru_utime;
 double starttime =(double)tim.tv_sec + (double)tim.tv_usec / 1000000.0;

 CLIQUES(interactions, mincliquesize, (void*)&params, insert_new_clique);
 printreport(clusters);
 if (clusters.size())
  filterclusters(clusters, index, interactions);
 printreport(clusters);

// return 0;

 // find partial cliques
 int count = clusters.size();

 for(float i=1; i>=ICDthreshold; i-=0.1)
 {
  while(findpartialcliques(clusters, interactions, i))
  {
   findpartialcliques(clusters, interactions, i);
   filterclusters(clusters, index, interactions); 
   printreport(clusters);
   if (clusters.size() == count) break;
   count = clusters.size();
  }
 }
/*
 while(findpartialcliques(clusters, interactions, ICDthreshold))
 {
  filterclusters(clusters, index, interactions);
  if (clusters.size() == count) break;
  count = clusters.size();
 }
 printreport(clusters);
*/
 getrusage(RUSAGE_SELF, &ru);
 tim = ru.ru_utime;
 double endtime =(double)tim.tv_sec + (double)tim.tv_usec / 1000000.0;
 cout << "Total time taken: " << (endtime - starttime) << endl;

 int i = 0;
 list<pair<int, float> > sortedclusters;
 for(int_set_vector::iterator it = clusters.begin(); it != clusters.end(); it++)
 {
  pair<int, float> fsweight = getFSWeight(*it, interactions);
  float avgfsweight = fsweight.first?(fsweight.second/fsweight.first):0;
  sortedclusters.push_back(pair<int, float>(i++, avgfsweight));
 }
 sortedclusters.sort(clustersCmp);
 i = 0;
 for(list<pair<int, float> >::iterator it = sortedclusters.begin(); it != sortedclusters.end(); it++)
 {
  float fsweight = it->second;
  set<int> &cluster = clusters[it->first];

  cout << "Cluster #" << i++ << ". Size " << cluster.size() << ". (FSWeight=" << fsweight << ")" << endl;
  int j = 0;
  for(set<int>::iterator ct = cluster.begin(); ct != cluster.end(); ct++)
  {
   if (j++ > 0) cout << " ";
   cout << protnames.at(*ct) << "(-)";
  }
  cout << endl << "----------" << endl;
 }
}
